﻿using System;

namespace Lambda
{
    class Program
    {
        delegate Fun Fun(Fun arg);

        private static int counter = 0;

        static void Main(string[] args)
        {
            // -------------------------------------
            // Church numerals

            Fun zero =  f => x => x;
            Fun one =   f => x => f(x);
            Fun two =   f => x => f(f(x));
            Fun three = f => x => f(f(f(x)));
            Fun four =  f => x => f(f(f(f(x))));
            Fun five =  f => x => f(f(f(f(f(x)))));
            Fun six =   f => x => f(f(f(f(f(f(x))))));
            Fun seven = f => x => f(f(f(f(f(f(f(x)))))));
            Fun eight = f => x => f(f(f(f(f(f(f(f(x))))))));
            Fun nine =  f => x => f(f(f(f(f(f(f(f(f(x)))))))));

            // -------------------------------------
            // Arithmetic with Church numerals

            Fun successor =   n => f => x => f(n(f)(x));
            Fun predecessor = n => f => x => n(g => h => h(g(f))) ((Fun)(u => x))(u => u);

            Fun subtract =    m => n => n(predecessor)(m);
            Fun add =         m => n => f => x => m(f)(n(f)(x));
            Fun multiply =    m => n => f => m(n(f));
            Fun exp =         m => n => n(m);

            // -------------------------------------
            // Boolean logic

            Fun @true =  a => b => a;
            Fun @false = a => b => b;

            Fun and =    p => q => p(q)(p);
            Fun or =     p => q => p(p)(q);
            Fun not =    p => p(@false)(@true);

            // -------------------------------------
            // Comparisons

            Fun isZero = n => n(x => @false)(@true);
            // Less-than-or-equal
            Fun leq =    m => n => isZero(subtract(m)(n));
            // Equals
            Fun eq =     m => n => and(leq(m)(n))(leq(n)(m));

            // -------------------------------------
            // Curry's Y-combinator

            Fun yc = f =>
                ((Fun)(x => f(x(x))))
                (y => f(a => y(b => y(b))(a)));

            // -------------------------------------
            // Fibonacci function

            Fun fib = f => n =>

                leq(n)(one) // n <= 1 ?

                (x => n(x)) // return n
                
                (x => (add(f(predecessor(n)))(f(predecessor(predecessor(n)))))(x));
                            // else return f(n-1) + f(n - 1 - 1);


            // -------------------------------------
            // Church-to-C# helper functions

            // int to Church numeral
            Func<int, Fun> cn = i =>
            {
                Fun result = zero;
                for (int j = 0; j < i; j++)
                    result = successor(result);
                return result;
            };

            // Church numeral to int
            Func<Fun, int> valn = n =>
            {
                counter = 0;
                n(x => { counter++; return x; })(null);
                return counter;
            };

            // Church predicate to bool
            Func<Fun, bool> valb = p => p(@true)(@false) == @true;

            // -------------------------------------
            // Some simple tests/demonstrations

            Console.WriteLine("Decoding church numeral seven:");
            Console.WriteLine(valn(seven));
            Console.WriteLine("(true and false) or (true and true) is:");
            Console.WriteLine(valb(or(and(@true)(@false))(and(@true)(@true))));
            Console.WriteLine("The 6th Fibonacci number is:");
            Console.WriteLine(valn(yc(fib)(six)));
            Console.WriteLine("The 25th Fibonacci number is (this could take some time):");
            Console.WriteLine(valn(yc(fib)(cn(25))));
            Console.WriteLine("The 25th Fibonacci number calculated without Church numerals:");
            Console.WriteLine(Fib(25));
            Console.WriteLine("Press enter to exit...");
            Console.ReadLine();
        }

        private static int Fib(int n)
        {
            if (n <= 1)
                return n;
            else
                return Fib(n - 1) + Fib(n - 2);
        }
    }
}
